// zero-assembler  for s-code  1 Sept 2007
//   change to relative jmp    7 Sept 2007
//   adapt to use in compiler class with somv2x 12 Sept 2012

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

typedef char name[32];

#define MAXSYM  500			// size of symbol table
#define MAXIN	10000		// size of input file

#define  eqs(a,b)   strcmp(a,b) == 0

char inbuf[MAXIN+10];
name sym[MAXSYM], tokstring, head, tail;
int value[MAXSYM], type[MAXSYM], freesym = 0;
int CH, inp = 0, line = 1;
int M[1000], CP = 1;

// type 1 op, 2 def sym, 3 label
// som-v24 instruction set
name opstr[] = {
	"nop","add","sub", "mul", "div","band","bor","bxor","not","eq",
	"ne", "lt", "le",  "ge",  "gt", "shl", "shr","mod", "ldx","stx",
	"ret","nop","array","end","get","put", "ld", "st",  "jmp","jt",
	"jf", "lit","call","nop", "inc","dec", "sys","case","fun","nop",
	"ads",
	// 41.. extra instructions
//	"ldi","sti","ldy", "sty","push","pop", "go", "ja",
	0
};

void out(int op, int arg){
	if( op >= 28 && op <= 30 ) arg -= CP; // jmp,jt,jf relative
	M[CP] = (arg << 8) | op;
	CP++;
}

int search(char *s){
	int i = 0;
	while(sym[i][0] != 0){
		if( eqs(s,sym[i]) ) break;
		i++;
	}
	return i;
}

int insert(char *s){
	int i;
	i = search(s);
	if( i == freesym ){
		strcpy(sym[i],s);
		freesym++;
	}
	if(freesym >= MAXSYM) printf("symbol table full\n");
	return i;
}

void getC(void){
	CH = inbuf[inp];
	if(CH != 0) inp++;	// not EOF
}

void goToEol(void){
	do
		getC();
	while( CH != 10 && CH != 0 );	// nl EOF
}

// getc until found, skip comment
void skipblank(void){
	for(;;){
		getC();
		while( isspace(CH) ){
			if( CH == 10 ) line++;
			getC();
		}
		if( CH == ';' ){
			goToEol();
			line++;
		}else
			break;
	}
}

void token(void){
	int k;
	skipblank();
	tokstring[0] = 0;
	if(CH == 0) return;	// EOF
	k = 0;
	while( ! isspace(CH) ){
		tokstring[k] = CH;
		k++;
		getC();
	}
	tokstring[k] = 0;
//	printf("'%s ",tokstring);
}

void dumpsym(void){
	int i = 0;
	while(i < freesym){
		if(type[i] != 1)
			printf("%s %d\n",sym[i],value[i]);
		i++;
	}
}

void readsym(void){
	int nm;
	token();
	while( ! eqs(tokstring,"#") ){
		nm = insert(tokstring);
		token();
		value[nm] = atoi(tokstring);
		type[nm] = 2;
		token();
	}
}

void asm(void){
	int nm, n, op, arg;
	token();
	while( ! eqs(tokstring,"#")){
		if( tokstring[0] == ':' ){			// label
			nm = insert(tokstring+1);
			value[nm] = CP;
			type[nm] = 3;
		}else if( isdigit(tokstring[0])){	// number
			out(atoi(tokstring),0);
		}else if( eqs(tokstring,".")){
			out(CP,0);
		}else if( eqs(tokstring,"..")){
			out(CP+1,0);
		}else{			// search op and sym
			n = strcspn(tokstring,".");
			strncpy(head,tokstring,n);
			head[n] = 0;
			if(n == strlen(tokstring))		// no arg
				tail[0] = 0;
			else
				strcpy(tail,tokstring+n+1);
			nm = insert(head);
			switch(type[nm]){
			case 0: out(0,0); break;		// undef
			case 1:							// op
				op = value[nm];
				arg = 0;
				if(tail[0] != 0)
					if( isdigit(tail[0]) )
						arg = atoi(tail);
					else
						arg = value[insert(tail)];
				out(op,arg);
				break;
			case 2:							// sym
			case 3: out(value[nm],0); break;
			}
		}
		token();
	}
}

void listing(void){
	int i, op, arg, k;
	char *s;
	for(i=1; i < CP; i++){
		op = M[i] & 255;
		arg = M[i] >> 8;
		for(k=0; k < freesym; k++)
			if ((type[k] == 1) && (value[k] == op))
				break;
		s = (k == freesym) ? "x" : sym[k];	// check op
		if(M[i] == i+1)
			printf("%d ..     %d\n",i,M[i]);
		else
			printf("%d %s %d   %d\n",i,s,arg,M[i]);
	}
	dumpsym();
}

// init symbol table with op
void putOp(void){
	int nm, i = 0;
	while(opstr[i][0] != 0){
		nm = insert(opstr[i]);
		value[nm] = i;
		type[nm] = 1;
		i++;
	}
}
// out object to a.obj
void outobj(void){
	int i;
	FILE *fo;
	fo = fopen("a.obj","w");
	fprintf(fo,"5678920\n1 %d\n",CP-1);
	for(i=1; i<CP; i++){
		fprintf(fo,"%d ",M[i]);
		if( i % 10 == 0 ) fprintf(fo,"\n");
	}
	fprintf(fo,"\n0 0\n"); // no data segment
	fclose(fo);
}


int main(void){
	int n;
	putOp();
	n = fread(inbuf,1,MAXIN,stdin);
	inbuf[n] = 0;
	readsym();
	n = inp;
	asm();
	inp = n;	// rewind
	CP = 1;
	asm();
	listing();
	outobj();
//	dumpsym();
}

/* -------------  test --------

void testop(void){
	int i;
	i = 0;
	while(op[i][0] != 0){
		printf("%s ",op[i]);
		i++;
	}
}

	int nm, i;
//	testop();
//	nm = insert("aa");
//	nm = insert("bb");
//	nm = insert("aa");
//	printf("%s\n",sym[nm]);
//	printf("%d\n",free);
//	printf("%d ",isspace(EOF));

int main(void){
	token();
	while( ! eqs(tokstring,"#") ){
		printf("%s %d ",tokstring, line);
		token();
	}
}

*/

